#include "chatManager.as"
#include "terrainEditor.as"
#include "gameScriptWrapper.as"
#include "welcome.as"
#include "streamValidation.as"
#include "botAPI.as"
#include "kickNotificationWrapper.as"

// Create a chatManager object instance
chatManager chatSystem();



// Create a gameScriptManager object instance.
gameScriptManager gsm();



// Add the autoModerator plugin.
// This plugin allows the following commands:
//     - !suspendchat   Suspend all chat traffic for a certain time
//     - !schat         Short for !suspendchat
//     - !chatban       Blocks all chat messages from a certain user for a
//                      certain time
//     - !cban          Short for !chatban
//     - !temporaryban  Not yet implemented
//     - !tempban       Short for !temporaryban
//     - !tban       Short for !temporaryban
// Furthermore, it adds the profanity filter and flood filter to the chat.
chatPlugin_autoModerator autoModerator(@chatSystem);

// This small plugin transform messages like:
//      Anonymous: OMG LOOK ATZ ME
// to:
//      Anonymous: Omg look atz me
//chatPlugin_disallowAllUpperCase disallowAllUpperCase(@chatSystem);

// This plugin adds the following commands:
//     - !afk           Prints "<username> is now away from keyboard."
//     - !back          Prints "<username> is now back."
//     - !gtg           Prints "<username> has got to go."
//     - !brb           Prints "<username> will be right back."
chatPlugin_miscellaneousCommands miscellaneousCommands(@chatSystem);

// This command adds the !move command, which will move your avatar in the
// specified direction.
chatPlugin_moveCommand moveCommand(@chatSystem);

// This commands adds the !goto, !setgoto (!sgoto) and !delelegoto (!dgoto)
// command. It allows you to store the position of your avatar (!setgoto <name>)
// and later teleport to that position (!goto <name>).
// Note: The !setgoto command works only with RoR 0.39! (not with 0.38)
chatPlugin_gotoCommand gotoCommand(@chatSystem, @gsm);

// This keeps track of what terrains are being used on the server and gives a
// list when the !terrainlist (!tlist) command is used.
// Note: All clients with RoR 0.38 will be shown as 'unknown terrain'.
chatPlugin_terrainList terrainList(@chatSystem, @gsm);

// This checks the latency of all RoR 0.39 players every 20 seconds and responds
// to !ping commands.
chatPlugin_ping ping(@chatSystem, @gsm);

// Another example, to show how to add commands.
// This adds a command !boost that gives a temporary boost to the truck of
// the user.
customCommand boostCommand(@chatSystem, "boost", customCommand_boost);
void customCommand_boost(chatMessage@ cmsg)
{
	cmsg.privateGameCommand.boostCurrentTruck(2);
}
// Extra boost commands
customCommand boost2Command(@chatSystem, "boost2", customCommand_boost2);
void customCommand_boost2(chatMessage@ cmsg)
{
	cmsg.privateGameCommand.boostCurrentTruck(5);
}
customCommand boost3Command(@chatSystem, "boost3", customCommand_boost3);
void customCommand_boost3(chatMessage@ cmsg)
{
	cmsg.privateGameCommand.boostCurrentTruck(8);
}
customCommand boost4Command(@chatSystem, "boost4", customCommand_boost4);
void customCommand_boost4(chatMessage@ cmsg)
{
	cmsg.privateGameCommand.boostCurrentTruck(10);
}


// Add the terrain editor (RoR 0.39 only!)
// Commented out, doesn't work correctly yet.
//terrainEditor ternnditor(@chatSystem, @gsm);



// Load the terrain edits
terrainFileLoader terrain(@gsm, server.getServerTerrain());



// Add some commands that allow you to reload the terrain file, without having to restart the server.
customCommand reloadTerrainCommand(@chatSystem, "reloadterrain", customCommand_reloadTerrain, null, true, AUTH_ADMIN);
void customCommand_reloadTerrain(chatMessage@) { terrain.reload(); }

customCommand unloadTerrainCommand(@chatSystem, "unloadterrain", customCommand_unloadTerrain, null, true, AUTH_ADMIN);
void customCommand_unloadTerrain(chatMessage@) { terrain.unload(); }

customCommand loadTerrainCommand(@chatSystem, "loadterrain", customCommand_loadTerrain, null, true, AUTH_ADMIN);
void customCommand_loadTerrain(chatMessage@)   { terrain.load();   }

// This is a list of stream names that will not be broadcasted 
array<string>  streamFilterList =  
{ 
	"2004HummerH2.truck","camry.truck","m3e93_gold.truck","2010_Chevy_Camaro_SS.truck","chevysilveradonascar.truck","chevysilveradonascarsuper.truck","svtraptor.truck","RenoClio.truck","Ferarri_458_Spider.truck","bmw750i1.truck","bmw750i2.truck","bmw750i3.truck","chevysilverado2.truck","chevysilverado2mud.truck","chevysilverado2500.truck","chevysilverado2500lifted.truck","chevysilverado3500.truck","gavrilmv4police.truck","gavrilmv4PoliceInterceptor.truck","gavrilmv4rpolice.truck","gavrilmv4RUSSIANPOLICE.truck","gavrilmv4sunmarkedpolice.truck","BoxDodgeRamEXT222.truck","svtraptoroffroadv1.truck","svtraptoroffroadv2.truck","gavril_monster_energy.truck","gavril_monster_energy1.truck","spacewagon_lhd.truck","forester.truck","foc07.truck","nissanr34.truck","camry97.truck","cvev.truck","2115.truck","vaz2.truck","vol.truck","ferrari_458_spider.truck","chevelle.truck","MMCP-C.truck","reven.truck","merc.truck","lincoln.truck","koenigsegg.truck","funnytex.truck","funnytex2.truck","hummerh.truck","focus.truck","Supra92.truck","BeastyG2.truck","Mafioso.truck","DatSwaqqerStickerBomber.truck","TriggerFinger.truck","bcgtpm.truck","gavrilmv4Di1.truck","gavrilmv4Di12.truck","gavrilmv4_ret.truck","gavrilluxury.truck","gavrilmv4_n24.truck","gavrilmv4_nur.truck","gavrilmv4_xd1.truck","gavrilmv4r_POLICe.truck","pri.truck","Peugeot406.truck","1997HondaCivicType-R.truck","gazel.truck","gavrilgv3_DRAG.truck","Ferrari1989F40Competizione.truck","daytonahazzard.truck","ATV.truck","1983AudiSportquattro.truck","1990_BMW_M5_E34.truck","tt.truck","2012Mercedes-BenzS65AMG[W221].truck","Niva.truck","2114u.truck","buffalo.truck","buffaloP.truck","20.truck","tax.truck","sultan.truck","gavrilmv4_jdm.truck","chevy10blazer.truck","ICCE300Meshed_Beta2L.truck","ICCE300Meshed_Beta3AGV_1.truck","private1.truck","private2.truck","private3.truck","GCPS-2016BluebirdVision.truck","GCPS-2016BluebirdVision_white.truck","GCPS-2016BluebirdVision_y.truck","GCPS-2016BluebirdVision_wy.truck","GCPS-2016BluebirdVision_s.truck","GCPS-2016BluebirdVision_white_S.truck","GCPS-2016BluebirdVision_wy_s.truck","GCPS-2016BluebirdVision_y_s.truck","JH-AmtranB800.truck","JH-Amtran3800.truck","JH-Amtran3800SED.truck","JH-AmtranB800SED.truck","1999ThomasMVPERFPS.truck","2007ThomasC2FPS.truck","2007ThomasC2.truck","2009BBA3RE.truck","2010ICCE.truck","2010ICCEFPS.truck","2013BBT3FE.truck","2013ThomasEFX.truck","2013ThomasEFXFPS.truck","2011BBA3.truck","2000BBFS65.truck","2001BB3800.truck","2006BBICCE.truck","moskvich412.truck","02CrownVicPoliceUnmarked.truck","06chargerunmarked.truck","12policecharger.truck","2011FordTarusSlicktop.truck","06policeimpala.truck","2011FordTaruPolice.truck","C2(40).truck","8552008BBVPBCRET.truck","dcds.truck","impala.truck","daytona.truck","AMC_Javelin2.truck","amcpacer.truck","ae86_compom.truck","silvias13.truck","silvias13DriftTune.truck","honda97.truck","r33.truck","subaru22.truck","2001su.truck","rx7.truck","nissanR34.truck","2010Shelbygt500.truck","lexusis300.truck","hondansxgt.truck","su05.truck","su05-2.truck","chevyimpala.truck","chevyimpalanowheels.truck","f430_scu.truck","2008su.truck","m3e92.truck","m3e92_gold.truck","2009DodgeChallengerSRT8(fps).truck","2009DodgeChallengerSRT8.truck","jeep_.truck","mazda3.truck","2010ChevyCamaroSS.truck","x5.truck","vwtour.truck","11'porsche911gt3.truck","Buggati.truck","suzuki.truck","12Mazda_RX8.truck","Dodge_ch_392.truck","moskvich.truck","2107.truck","300ZX.truck","350Z.truck","2104.truck","115.truck","passat.truck","bcgtpm2.truck","continental.truck","m3.truck","e38.truck","cater500.truck","Chevroletss.truck","dmc12.truck","vip.truck","vip2.truck","vip3.truck","fer.truck","ToyotaSoarer.truck","Fordfiesta.truck","gavrilmv2r.truck","gavrilgv5_RA1PID.truck","delsol.truck","kalina_hatch.truck","ccx.truck","land200.truck","Maibatsu_Sport.truck","Maibatsu_LT.truck","Maibatsu.truck","rx7drift!.truck","mazda-rx-7-mad-mike.truck","3dc2eUID-MM-MagmaEdition-400-ls.truck","evoix.truck","evox2.truck","evox.truck","nissan240sx3.truck","nissan240sx3_dragster.truck","nissan240sx3_drifter.truck","nissan240sx3_fps.truck","nissan240sx3_drifter_fps.truck","Primera.truck","Nissan_GTR-R35.truck","nislaubos.truck","cuda.truck","R32.truck","rx7drifttune.truck","brzs.truck","brzs_blue.truck","supra.truck","supradrifttune.truck","Supra.truck","Supra2.truck","soarer.truck","21099.truck","Volvo242.truck","Volvo242drag.truck","Volvo242Street.truck","Volvo242WRC.truck","Race_Touareg.truck","2009-2.truck","2010-2.truck","2011-2.truck","2012-2.truck","2005-2.truck","2006-2.truck","2007-2.truck","2008-2.truck","2013-2.truck","2013-3.truck","2013-4.truck","2014-2.truck","2015-2.truck","2016-2.truck","2017-2.truck","2018-2.truck","2019-2.truck","NSCbus1.truck","NSCbus34.truck","NSCbus42.truck","NSCbus9.truck","NSCbus93.truck","NSCbus94.truck","NSCbus36.truck","NSCbus32.truck","NSCbus59.truck","ICCE300Meshed_Beta2K.truck","ICCE300Meshed_Beta2K.TRUCK","2007-IC-CE-GCSNC-Bus-24.truck","2010ICCE-N.truck","2006ICCE-A.truck","2017ThomasC2-SEMINOLE-61615.truck","JPS2018-181.truck","LABS9580.truck","2014ICFE300.truck","1992BBB700-0392.truck","lacheti.truck","2019BBV24.truck","2018-TRRS-281.truck","2017BBV25.truck","2012BBV7.truck","1992BBC60-0292.truck","2018QISD1817.truck","2005QISD0523.truck","2003QISD0318.truck","2004-IC-CE-GCSNC-Bus-1.truck","2005ICCE200-HOLCOMB-137.truck","af25UID-aaadodgeviper.truck","CSMVPER.truck","CSMVPR2.truck","NBBC500-PUBLICR-5_WIM-I.truck","1998ThomasVista[I].truck","14-21.truck","2017ThomasHDX-SEMINOLE-61711.truck","2008ICCE300-179-SH-Lucini.truck","GSX1.0.truck","2008C2J.truck","2010C2J.truck","2010C2A.truck","2014C2D.truck","2016C2N.truck","2006-IC-CE-GCSNC-Bus-22.truck","55-2015BBIntWorkstar.truck","2011ICCE300-3506.truck","2017ICCE300-3892.truck","2006FS65-24.truck","gavril_r_gold_legend.truck","2009-Thomas-C2-NC.truck","2001-Thomas-FS65.truck","2014ThomasEFX.truck","2015ThomasEFX.truck","2016ThomasEFX.truck","2017ThomasEFX.truck","2018ThomasEFX.truck","2004ICCE-Lucini-183.truck","1969DodgeChargerRT.truck","79ChevyCamaroZ28.truck","54a9UID-2006Charger-Red.truck","54a9UID-2006ChargerADE.truck","54a9UID-2006ChargerDET.truck","54a9UID-2006ChargerLAPDST.truck","54a9UID-2006ChargerPPB.truck","54a9UID-2006Charger-Blue.truck","54a9UID-2006Charger-Black.truck","54a9UID-2006Charger-Silver.truck","1949_Ford_Gasser.truck","2010Shelbygt500ERCAWD.truck","2010Shelbygt500ERCRWD.truck","C10lololol4.truck","evoix_saquet.truck","1979_Blazer_ERC.truck","bel.truck","ffrskb.truck","2004BB3800-SERSD-250.truck","2004BB3800-SERSD-265.truck","2015BBV_SERSD-12.truck","AastonMartinVanquishf.truck","diablo.truck","2005-449.truck","C2(FS-5).truck","2004QISD0440.truck","2016-IC-CE-GCSNC-Bus-16.truck","2006-IC-CEGCSNC-Bus-19.truck","1999GCPSTC2000.truck","2012-IC-CE-Bus-20.truck","2006ICCE-Ds.truck","2006ICCE-Jn.truck","2007ICCE-D.truck","2008ICCE-JD.truck","2011ICCE-E.truck","2015ICCE-C.truck","2014ICCEACT.truck","1993-Ford-B700.truck","99-02.truck","2010ThomasC2NewYork.truck","2010ThomasC2NewYorkFPS.truck","2018-260.truck","2017QISD1739.truck","99-01.truck","2009ThomasC2ED706.truck","2016ThomasC2ED944.truck","1705.truck","2020ThomasC2CarolinaThomas.truck","2006C2141.truck","2006C2153.truck","2009C2299.truck","2006C2157.truck","2003QISD0321.truck","2019ICCE-ED41.truck","2013ICCE-ED26H.truck","2020ICCE-ED53.truck","audi90.truck","2004-IC-CE-GCSNC-Bus-7.truck","2010-IC-CE-GCSNC-Bus-9.truck","2006ICCEGCSNCBus17.truck","Bus25-04.truck","2015ICCE-1515ED.truck","Bus20-16.truck","2015ICCE300.truck","Bus19-10.truck","Bus7-12.truck","Bus50-08.truck","AstonMartin_v8_Vantage.truck","SheriffTaurus.truck","PoliceUndercoverTaurus.truck","PoliceTaurus.truck","WhiteTaurus.truck","OrangeTaurus.truck","NycCab.truck","Barwood.truck","GreyTaurus.truck","GreenTaurus.truck","CivTaurus.truck","BlueTaurus.truck","BlackTaurus.truck","crownviccop1.truck","Bus9-08.truck","Bus30-14.truck","Badass_Bandit.truck","Track_Bandit.truck","Street_Bandit.truck","pri_gold.truck","TCPSD60.truck","Lucini-CCS2919.truck","Lucini-MJ1772.truck","Lucini206.truck","DTC997.truck","1986-GMC-Diesel-6000.truck","1999-Freightliner-FS65.truck","2001ToyotaChaser.truck","08_C2_L.truck","2017DSS0321.truck","2017DSS1742.truck","2017DSS1743.truck","HECK-MEYERS2009ThomasC2191.truck","HECK-MEYERS2010ThomasC2165.truck","HECK-MEYERS2919.truck","HECK-MEYERS315.truck","Bus21.truck","2003ICCE-K867.truck","2018Handi-283.truck","1991MazdaRx7.truck","2010-Bluebird-All-American.truck","2010_BB_AA_RE_NC.truck","HECK-MEYERS947.truck","2015-572-P.truck","burnside_badass_convertible.truck","NoDIstrict7.truck","NBBC400-Baldi2021.truck","Bald'sBus.truck","TONAWANDA2015CE300SPARE-47.truck","bluebirdbus.truck","2000AmTranREBusHeritage66(HI)(NWSA).truck","NBBCtest.truck","activity23.truck","activity24.truck","SEVS2003BBV.truck","SEVS2004BBV.truck","SEVS2005BBV.truck","SEVS2006BBV.truck","SEVS2007BBV.truck","SEVS2008BBV.truck","SEVS2009BBV.truck","SEVS2010BBV.truck","SEVS2011BBV.truck","SEVS2012BBV.truck","SEVS2013BBV.truck","SEVS2014BBV.truck","SEVS2015BBV.truck","SEVS2016BBV.truck","SEVS2017BBV.truck","SEVS2019BBV.truck","SEVS1996BBTC2000.truck","SEVS1997BBTC2000.truck","SEVS1997BBTC20009724.truck","SEVS1997BBTC20009725.truck","SEVS1998BBTC2000.truck","SEVS1999BBTC2000.truck","SEVS1999BBTC20009900.truck","SEVS1999BBTC20009940.truck","SEVS1999BB3800991.truck","SEVS2005ICFE300.truck","SEVS2006ICFE300.truck","SEVS2007ICFE300.truck","SEVS2008ICFE300.truck","SEVS2009ICFE300.truck","SEVS2010ICFE300.truck","37.truck","SEVS2005CE200.truck","SEVS2006CE200.truck","SEVS2009CE300.truck","SEVS2010CE300.truck","SEVS2012CE300.truck","SEVS2013CE300.truck","SEVS2019CE300.truck","KEVINSEVS2017ICCE.truck","CHRISSEVS2017ICCE.truck","RYANSEVS2017ICCE.truck","bus6.truck","bus7.truck","bus26.truck","bus27.truck","bus28.truck","bus29.truck","bus30.truck","bus31.truck","bus32.truck","bus33.truck","2009BBA3FE.truck","1994-International-3800-DT408.truck","2006-IC-CE200.truck","2010-IC-CE300.truck","Prison.truck","Shortie.truck","Coach.truck","2010-IC-CE200.truck","Lucini237.truck","1998ThomasVista[NS].truck","2010ThomasC2-2339.truck","2003ICCE-J.truck","2016ICCE-ED66.truck","2019ICCE-ED42H.truck","2018BBV-TRRSD-162.truck","2005-BMW-E46.truck","91ThomasERFivestartrans.truck","ThomasInt3800.truck","2006ICCE-E1H.truck","2018BBV-TRRSD-161.truck","2011ICCE-E801.truck","2004ICCE-E225.truck","2007ICCE-E643.truck","NCPS897.truck","2010ICCE-FL.truck","1103.truck","2016 Bus 154.truck","1997BBTC2000Bus111.truck","1999BBTC2000Bus118.truck","2004CROWN-FREIGHTLINER-TRRS-103.truck","cv-ICFE300Old.truck","109.truck","1503.truck","1997GCPS9818.truck","Bus9-13.truck","2005BBFS65.truck","2019TRRSD-259.truck","2006ICCE-E13.truck","2007BBV-1407.truck","2008BBV-1508.truck","svtstang.truck","2007-IC-CE-GCSNC-Bus-25.truck","87 Murcury Grand Marquis.truck","2018 HECK-MEYERS HDX (1412).truck","2018 HECK-MEYERS HDX(1407).truck","bus25.truck","2006BBSBCV-1306.truck","2005BBV-1205.truck","09'BBV95.truck","dodgecaravan.truck","2005  Lamborghini Gallardo TEST MODEL.truck","TCTC2008B4.truck","2002A3CCS0902.truck","2000A3CCS2800.truck","HLS2018B5.truck","HLS1999B27.truck","HLS1998B21.truck","2002A3CCS3502JH.truck","HLS2008B16.truck","HLS2007B4.truck","HLS2003B38.truck","2001A3CCS4801.truck","2001A3CCS4901.truck","2002A3CCS0802.truck","HLS1998B26.truck","2000A3CCS2500.truck","HLS1998B22.truck","HLS2008B34.truck","HLS2006B40.truck","2004A3CCS5804.truck","2001A3CCS7001.truck","TCTC2004B3.truck","2001A3CCS3701.truck","2003A3CCS2203.truck","2004A3CCS2704.truck","HLS2008B6.truck","HLS1998B25.truck","HLS2007B3.truck","HLS1999B12.truck","HLS2016B33.truck","2003A3CCS0403.truck","2003A3CCS1703.truck","2002A3CCS2602.truck","2002A3CCS2102.truck","HLS2008B2.truck","HLS2000B30.truck","HLS2003B32.truck","HLS2018B11.truck","HLS2006B41.truck","2000A3CCS3000.truck","HLS2001B7.truck","2002A3CCS3202.truck","2001A3CCS1401.truck","2000A3CCS3100.truck","HLS2008B35.truck","2004A3CCS4204.truck","2003A3CCS3403.truck","HLS2016B13.truck","HLS2000B14.truck","HLS2002B31.truck","HLS1998B19.truck","1997GCPSTC2000.truck","2003A3CCS3903.truck","2003A3CCS4403.truck","HLS2010B8.truck","2004A3CCS3304.truck","LLS1993B8.truck","LLS1997B13.truck","2003-IC-CE200.truck","Lucini207.truck","Lucini159.truck","Lucini238.truck","DTC525.truck","HUNTMAN 7.truck","Bus20-10.truck","2015ICCE-1315M.truck","cv-bluebird3800.truck","2016ICCE-E907.truck","2002BBFS65-1102.truck","2006ICCE.truck","2008ThomasC2ED653.truck","2016-122.truck","2018-983.truck","2015-IC-CE-GCSNC-Bus-4.truck","2008-IC-CE300.truck","2007-Thomas-C2-NC.truck","2003CE-3-Drag.truck","195LBL-WB2014T3.truck","Lucini193.truck","LLS2007B5.truck","LLS2019B4.truck","Lucini913.truck","Lucini274.truck","Lucini2215.truck","LuciniCCS0518.truck","23-2002IC.truck","1993_335.truck","LuciniCCS6114.truck","07_C2_L.truck","14_C2_L.truck","10_C2_L.truck","C-2005-51.truck","C-2014-03.truck","C-2008-18-H.truck","17_C2_L_SE.truck","vh-03fs65.truck","NC16CEL.truck","2005 ic ce personal f.t.a.g.truck","06impalaunmarked.truck","15 SLT MVP.truck","C2 105 CEO.truck","2015dmax4wd.truck","2015dmax4wdmudder.truck","2015dmax2wdslammed.truck","2015dmax4wdslammed.truck","2015dmax2wd.truck","2015dmax4wdcencalstyle.truck","2015dmax4wdflatbed.truck","NCBC18.truck","2015-IC-CE GCSNC-Bus-14.truck","hazzard.truck","79ChevyCamaroZ28E.truck","BlossomTransportation2011BlueBirdAllAmericanFeHandicappedAccess.truck","Lucini203.truck","Lucini-MJ1711.truck","Lucini208.truck","Lucini-CCS5718.truck","Lucini-CCS0315.truck","Lucini-CCS5211.truck","Lucini905.truck","Lucini507.truck","Lucini401.truck","2019TRRSD-262.truck","2006ICCEFPS.truck","2008ICFE-274.truck","TRRS-177.truck","2009ThomasC2ED6246.truck","2010ThomasC2ED6250.truck","2013ThomasC2ED6252.truck","2013ThomasC2ED6253.truck","2013ThomasC2ED6254.truck","2015ThomasC2ED6258.truck","2019ThomasC2EDCrewdz.truck","2017H-603.truck","2007-1.truck","2008-1.truck","2014-1.truck","2013-1.truck","2003-432.truck","2009-1.truck","2020.truck","2011-1.truck","2005-1.truck","2019-1.truck","2010-1.truck","2012-1.truck","2018-1.truck","2017-1.truck","2016-1.truck","2016-590.truck","2014-565.truck","2015H-578.truck","2011H-519.truck","2006-1.truck","2015-1.truck","2018-18-5.truck","NCBC21.truck","2017ICCE-MT.truck","08policetahoe.truck","2019TRRSD-263.truck","2019TRRSD-261.truck","FS2019ICCEHX8945.truck","FS2019ICCEHX8946.truck","FS2020ICCEHX8844.truck","2019BBVFSHD4566.truck","2019BBVFSHD4568.truck","995BBGMC-0495.truck","1998BB3800-0698.truck","2002BBFS65-1002.truck","LABS0836.truck","2007ICCE200-STOUGHTON-071232.truck","2016ThomasC2-STOUGHTON-8639.truck","Lucini205.truck","2004CROWN-FREIGHTLINER-TRRS-108.truck","FST2010C2 10-07.truck","FST2018C2 10-01.truck","2017Activity(Universal Shuttle).truck","2018Handi-671.truck","2015 IC CE 300.truck","2015QISD1505.truck","2021QISD2130.truck","2006QISD0633.truck","2010QISD1021.truck","2020QISD2027.truck","2015QISD1526.truck","2005QISD0508.truck","Lucini194.truck","LuciniCCS2419.truck","Lucini275.truck","LuciniM&J2851.truck","2018BBV-165.truck","2007 C2 162.truck","2017 C2 166.truck","2018Handi-321.truck","Lucini192.truck","DTC506.truck","Lucini204.truck","AGRP2014CE.truck","2018BBV-TRRSD-168.truck","TRRS-116.truck","2010ICCE-ED22.truck","2019TRRSD-258.truck","2002-2.truck","gavrilmz2rGT1Ravenwest.truck","gavrilmz2rGT1.truck","2016-IC-CE-GCSNC-Bus-13.truck","Hauck38.truck","11-04.truck","2010ThomasC2Handi.truck","2011ThomasHDX-Serengeti-Bus60.truck","bmw-e39-m5-2004.truck","vh-01fs65.truck","NCPSSfORD_7.truck","2007DSS0730.truck","CBMNCBUS1.truck","vh-01fs65101.truck","2006QISD0632.truck","2007BluebirdVision.truck","TCPSD162.truck","TCPSD28.truck","406.truck","8017.truck","vh-GMCThomas.truck","vh-05fs65.truck","3800D267.truck","8030.truck","vh-b800.truck","2010C2308.truck","Bus982.truck","2004FS65-SERSD-UN.truck","2006FS65-SERSD-17.truck","2010C2325.truck","2019C2560.truck","14'BBV95.truck","2002ICSB0215.truck","2012-IC-CE GCSNC-Bus-3.truck","2009-498.truck","CCS2800.truck","2017DSS1739.truck","1998BB3800-0798.truck","1995BBGMC-0495.truck","CCS3608.truck","2018ICCE300-3922.truck","2003QISD0319.truck","APS2.truck","APS9.truck","CCS2200.truck","2002BBChevy-0802.truck","APS16.truck","CV200.truck","2016-IC-CE-GCSNC-Bus-11.truck","2006-IC-CE GCSNC-Bus-20.truck","TRRS-135.truck","APS11.truck","TRRS-125.truck","2017BBV-RVCS-6902.truck","55-2015BBIntWorkstar_y_w_roof.truck","MPSBUS8.truck","APS12.truck","2005QISD0506.truck","2005QISD0522.truck","2010QISD1004.truck","2008QISD0810.truck","2020ThomasC2ED868.truck","2002-132.truck","8047.truck","kitt.truck","logan.truck","1986-Ford-B700.truck","1981-Chevrolet-V8-60.truck","1998-Freightliner-FS65.truck","01-FS65.truck","2017QISD1705.truck","2021ThomasC2-SEMINOLE-J.truck","2016ThomasC2-SEMINOLE-61521.truck","2016ThomasC2-SEMINOLEH.truck","2019ICCE-SEMINOLE-61904.truck","2018ThomasHDX-SEMINOLE-61708.truck","2019ICCE-SEMINOLE-61905.truck","2019ThomasHDX-SEMINOLE-61810.truck","TRRS-159.truck","2018BBV-SEMINOLE-61704.truck","2017QISD1707.truck","APS13.truck","CCS1506.truck","1969DodgeChargerRTStreetDrag.truck","vh-05fs6527.truck","2015-IC-CE-GCSNC-Bus-5.truck","APS6.truck","2018ICCE-ED997.truck","2004FS65-28.truck","3800D200.truck","vh-99fs651141.truck","ZAZtavria.truck","2101.truck","2019ICCE-SEMINOLE-61901.truck","2019ICCE-SEMINOLE-61902.truck","2019ICCE-SEMINOLE-61906.truck","2019ICCE-SEMINOLE-61908.truck","2019ICCE-SEMINOLE-61907.truck","2020ThomasC2-PASCO COUNTY SCHOOLS-62001.truck","burnsidelimolowerfaster.truck","2012ICCE300-3694.truck","2012ICCE300-3685.truck","2011ICCE300-3666.truck","11.truck","m3e42.truck","veyron.truck","1995-International-3800-DT466.truck","ZAZ.truck","gavril_r_sakura_bunny.truck","ae86_trueno_1st-3rd.truck","ae86_trueno_4th-6th.truck","Thomas-C2.truck","2016ThomasC2ED.truck","Murcielago2.truck","2021ThomasHDX-SEMINOLE-62006.truck","5-2015BBIntWorkstar_y_y_roof.truck","55-2015BBIntWorkstar_b_y_roof.truck","1999 Thomas Vista.truck","08_C2_LS.truck","2011BBD3FEhandy.truck","2011BBD3RE.truck","2011BBD3Church.truck","04-Thomas-FS65.truck","2004-IC-RE-RB.truck","BakuganDragonoidBreakable.truck","BakuganDragonoidIceBreakable.truck","TCPSD23.truck","TCPSD25.truck","BakuganDragonoid2019.truck","IceBakuganDragonoid2019.truck","BakuganDragonoidMurphy2019.truck","BakuganDragonoid2019Aurelus.truck","BakuganMatteBlackSexyEdition.truck","1995BBGMC-0595.truck","1989BBS1800-0189.truck","Lucini201.truck","006ICCE-J174.truck","2004ICCE-G135.truck","CCSD190519.truck","2010ThomasC2ED753.truck","2007C2SLTI50.truck","2020C2SLTI51.truck","2013ICCE-ED12.truck","2018ICCE-ED14.truck","2018TRRS-283.truck","2020ThomasC2ED420.truck","2009ICCE300-3462.truck","PCS Bus 17.truck","PCS Bus 14.truck","2014ICCEFPS.truck","ICTESTNS.truck","NCPS176.truck","2004ICCE200248.truck","FS65R1-1998ThomasMVPEF.truck","2010ICCE-C10.truck","gavril_r_rhd.truck","2009ThomasC2ED751.truck","DTC837.truck","2019QISD1902.truck","DTC556.truck","DTC817.truck","DTC917.truck","DTC847.truck","DTC546.truck","2015_DSS_1526.truck","2006-ICCE-GCSNC-Bus-8.truck","urvan.truck","vick.truck","Collins785.truck","BakuganDraganoid2019.truck","2005SBCV-ST32.truck","NCSFC 191.truck","NCS Forsyth County 3001.truck","DSS525.truck","DSS556.truck","2021QISD2128.truck","2013ThomasC2ED82.truck","2018ThomasC2ED10.truck","2015ThomasC2ED8069.truck","55-2015BBIntWorkstar_I.truck","fprd_tr.truck","55-2015BBIntWorkstar_y_y_roof.truck","55-2015BBIntWorkstarFS.truck","55-2015BBIntWorkstar_S2.truck","BBV1005.truck","BBV997.truck","BBV963.truck","BBV956.truck","BBV960.truck","2016BBV5.truck","2016BBV7.truck","2007BBARE_W_Y.truck","2005ICCEHandi-SEMINOLE-70329.truck","2017BBV-SEMINOLE-61705.truck","2017BBV-SEMINOLE-61703.truck","2005ICCEHandi-SEMINOLE-70325.truck","NCTrans2003ThomasFS65.truck","2016ThomasC2-PASCO COUNTY SCHOOLS-61515.truck","DSS197.truck","20030327.truck","2002-1.truck","DURHAM2013ICCE300.truck","CDS200789141.truck","2017QISD1726.truck","CDS2008157.truck","QISDVic.truck","DSS439.truck","DSS476.truck","2003SEVS0314.truck","21-92.truck","21-1.truck","2007FS65.truck","20030316.truck","20030314.truck","2005CE200-180.truck","chevy10blazer2.truck","3800J569.truck","2020ICCE-ED1002.truck","2018BBV-TRRSD-172.truck","Bus3-08.truck","Buffalo03ICCE208433.truck","EBPS9.truck","2018BBV19.truck","2005QISD0527.truck","2019ICCE300P.truck","FST2012C2 10-08.truck","FST2010C2 10-14.truck","2017BBV_FirstStudent-86.truck","2016BBV_FirstStudent-87.truck","FS8051.truck","FST2015C2 10-18.truck","2003FS65-26.truck","66.truck","85.truck","88.truck","2010ICCE300-3492.truck","21-90.truck","21-132.truck","2018-612.truck","2015-576.truck","2008-489.truck","TRRS 2017 Thomas C2.truck","2017BBV-RVCS-6609.truck","2012ICCE300-3456.truck","20020215.truck","DSS817.truck","2006CE200-181.truck","87.truck","TRRS177.truck","FST2016C2 10-04.truck","FST2016C2 10-09.truck","3800D264.truck","2002ThomasMVPER-172.truck","2016ThomasC2-SEMINOLE-61529.truck","TCPSD79.truck","2018TRRS-212.truck","2014_C2_SLTI_13.truck","MPSBus54.truck","2000-Corbeil-FSC.truck","2001-Corbeil-shorty.truck","2001-Corbeil-DRAG.truck","2000-Corbeil-highheadroom.truck","2020transbus2056.truck","CLSD2015CE300.truck","ASD_2014ThomasEFX-140180.truck","MPSBUS66.truck","MPSBus55.truck","MPSBUS65.truck","86.truck","2012ThomasC235.truck","MPSBUS2.truck","HECK MEYERS 20.truck","DeshaunC2.truck","LABS1254.truck","LABS1590.truck","LABS1661.truck","LABS1663.truck","LABS1589.truck","2005DSS0529.truck","2006ICCE-J.truck","ACXcustom.TRUCK","74.truck","58.truck","67.truck","59.truck","2002DSS0215.truck","UPS.truck","HondaPreludeWhite.truck","gallardotest.TRUCK","CCS4710.truck","2007SCPSC2.truck","1234.truck","84.truck","PB-37-18.truck","APS8.truck","11' porsche 911gt3.truck","2002ThomasMVPER-173.truck","EBPS1.truck","EBPS25.truck","TPS4.truck","APS15.truck","TPS3.truck","APS 30.truck","APS26.truck","APS10.truck","MTI807.truck","2017Activity.truck","2017ICCE300-HOLCOMB-2109.truck","Blue_Bird_Vision_Activity_Bus_CT.truck","2020SCSD4.truck","2016ThomasC2-BCS-932.truck","2021QISD2129.truck","LABS1255.truck","2018QISD1820.truck","93_chevy_silverado.truck","PCS Bus 12.truck","2019FSC2-114.truck","2019FSC2-117.truck","21 SLT HDX.truck","2018C2-MPS-18.truck","2018C2-MPS-27.truck","3.truck","UPT2015BBTX4FE.truck","2011ICCE-MPS-32-H.truck","1999 MVP 15.truck","2016ThomasC2-SEMINOLE-61502.truck","RidleySchoolDistrictBus58.truck","Deluxe_Bandit.truck","Bus24-06.truck","2006FS65-UN2.truck","1997LUCINI145.truck","2002ThomasMVPER-176.truck","CCS4110.truck","WSFCS2013ThomasC2Bus500.truck","2015.truck","2010.truck","2008.truck","2005.truck","2018.truck","2016.truck","2006.truck","2017ICCE300-125.truck","2009.truck","2014.truck","2013-IC-CE-Bus-92.truck","ThomasC2Bus90.truck","2006-IC-CE-GCSNC-Bus-21.truck","SEVS95_C2_IM.truck","WSFCS 1990 3800 Activity Bus830.truck","WSFCS 1991 3800 Activity Bus8047.truck","WSFCS19973800Bus273.truck","WSFCS19973800Bus271.truck","2020ThomasC2ED417.truck","WSFCS 2004ICCE 121.truck","WSFCS 2011ICCE 421.truck","15-39.truck","ICTEST.truck","2005QISD0534.truck","torino_flatout_2.truck","GCPS-02InternationalFE.truck","2011ICCE-E810.truck","2010_DSS_1021.truck","2021_DSS_2130.truck","2018BBV-SOUTHCAROLINA-0249.truck","2005HCPS0506.truck","2005HCPS0524.truck","2005HCPS0525.truck","2018HCPS1817.truck","2001QISD0138.truck","2018C2-SCPS-508-8888.truck","BakuganDraganoid2019Aurelus.truck","2006ICRE300EDIT.truck","2005QISD0509.truck","2017QISD1743.truck","2007QISD0728.truck","2017QISD1744.truck","2005QISD0524.truck","2007QISD0730.truck","2020LPS2020.truck","857.truck","907.truck","2017ICCE300-110.truck","2017ICCE300-114.truck","2017ICCE300-118.truck","2017BBV20.truck","2017ICCE300-115.truck","FirstStudent218696.truck","2017ICCE300-119.truck","HCL23.truck","2017BBV23.truck","2012ICCE300-MPS-43.truck","2010-IC-CE GCSNC-Bus-15.truck","NCPS810.truck","2010_DSS_1004.truck","104.truck","2021EISD2129.truck","2007FS65_DSS.truck","Phast Phantom.truck","burnsideHEMIv2.truck","FC3SturboRyosuke.truck","burnsideRallyC.truck","1012.truck","1014.truck","WSFCS Thomas GMC.truck","2017ICCE47.truck","2008ICBE60.truck","2001-ThomasMVPER.truck","2014QISD1413.truck","2014QISD1414.truck","2017QISD1722.truck","CSDS89BBS18008900.truck","DKSD 2008 ICCE 23.truck","chevelle72 Drag.truck","1968 c10 modest lift.truck","1969DodgeCharger RTTuned.truck","AMC_Javelin.truck","2009 Dodge Challenger SRT8 (fps).truck","2009 Dodge Challenger SRT8.truck","1969DodgeChargerRTOffroad.truck","burnsidelimolower.truck","bus24.truck","FREISE1855.truck","bus23.truck","CollinsNew2008ICCEACT.truck","CollinsNew2010ICCE.truck","CollinsNew2010ICCEFPS.truck","Collins2005NCICCE43.truck","Collins2005NCICCE51.truck","Collins2006NCICCE44.truck","Collins2007NCICCE50.truck","Collins2008NCICCE66.truck","Collins2008NCICCE596.truck","Collins2009NCICCE19.truck","Collins2011NCICCE800.truck","Collins2011NCICCE810.truck","Collins2015NCICCE109.truck","Collins2018NCICCE508.truck","Collins2018NCICCE510.truck","Collins2020NCICCE881.truck","08policetahoeslicktop.truck","02CrownVicPoliceALPR.truck","2006ICCE52.truck","SEVS95_C2_LMUH.truck","SEVS95_C2_L.truck","2007QISD0731.truck","Collins2011NCICCE830.truck","Collins2011NCICCE833.truck","Collins2011NCICCE823.truck","2020ThomasC2ED191.truck","Collins2020IC.truck","2018ICCE.truck","2019ICCE.truck","vh-1800JR.truck","2004-Dodge-Durango-diesel-police.truck","IceBakuganDraganoid2019.truck","CISD164.truck","WSFCS 2018ICCE 8941.truck","2018ThomasC2ED210.truck","05-FS65.truck","2004-IC-RE-T1.truck","2000short.truck","Bus8.truck","2018QISD1823.truck","2015- 1592.truck","807.truck","2019 - 904.truck","2020QISD2024.truck","1912.truck","2006QISD0631.truck","E36_PANZER.truck","APS14.truck","CCS3000.truck","accordsilvercomet.truck","2007QISD0713.truck","20-33.truck","135.truck","55-2001BBInt3800.truck","2010ThomasC2HandyBCS949.truck","02CrownVicPoliceUnmarkedCUSTOM.truck","2015ICCE-39.truck","TRRS-129.truck","2018BBV-TRRSD-166.truck","2006ICCE200-FSBUFFALO-201046.truck","UPT2002ICCE200420.truck","2008ICCE200-FSBUFFALO-201188.truck","2012ICCE300-FSBUFFALO-127643.truck","2007ICBE200-BUFFALO-207477.truck","2010ICCE200-FSBUFFALO-100102.truck","NCPS-2011BBD3RE.truck","SEVS95_C2_LMU.truck","SEVS95_C2_IH.truck","burnsideundercoverpolice.truck","WSFCS2016ThomasC2Bus801.truck","WSFCS19923800Bus569.truck","WSFCS 2004ICCE 122.truck","WSFCS 2018ICCE 430.truck","WSFCS2020ThomasC2Bus860.truck","WSFCS2020ThomasC2Bus861.truck","WSFCS 2007ICCE 55.truck","WSFCS 2016CCE 425.truck","WSFCS 2018ICCE 428.truck","WSFCS 2018ICCE 429.truck","2008ThomasC2ED651.truck","WSFCS2019ThomasC2Bus840.truck","TC2000.truck","Jason2019CE61901.truck","F350 offorad.truck","bus9.truck","CamdenTest2020.truck","CCS0518.truck","CCS1810.truck","CCS5211.truck","CCS6114.truck","CCS6214.truck","08_C2_LH.truck","FSBus43.truck","2006ICBE200-STOUGHTON-060872.truck","2006ICCE200-STOUGHTON-060480.truck","2011ICCE-STOUGHTON-H.truck","2016ThomasC2-STOUGHTON-8636.truck","2016ThomasC2-STOUGHTON-8645.truck","2016ThomasC2-STOUGHTON-8640.truck","2010ICCE300-STOUGHTON-101900.truck","FSBus31.truck","2020SwitzerCartyThomasC2.truck","2016ThomasC2-PASCO COUNTY SCHOOLS-61526.truck","2017ICSB1744.truck","2016ICCE-ST711.truck","2004BB3800-SEMINOLE-60315.truck","2004BB3800-SEMINOLE-60318.truck","2009BBA3RE-CCSD-324.truck","NorthWestTransportation2152.truck","2016ThomasC2-PASCO COUNTY SCHOOLS-61523.truck","2016ThomasC2-PASCO COUNTY SCHOOLS-61521.truck","TRRS-173.truck","MEVS_2009_ICCE300_Bus43v4.truck","2020M2CBMCE300.truck","AmTranRE.truck","MEVS_2009_ICCE300_Bus43.truck","MEVS_2009_ICCE300_Bus43v2.truck","2019NWBUSNBBC-5.truck","NBBC400-7.truck","2019NWBUSNBBC-6.truck","CBM2020Vision.truck","2006ICFE300-4.truck","2006ICFE300-2.truck","2006ICFE300-3.truck","2006ICRE300-1.truck","2006ICRE300-2.truck","SMFCS33.truck","2021CBMHDX.truck","2020M3CBMCE300.truck","cbm-3800.truck","2004MEVS51-8.truck","2004MEVS64-8.truck","2001BB3800CBM.truck","2004NCPSBC6.truck","2004NCPSBC7.truck","2004NCPSBC8.truck","2004NCPSBC9.truck","2004NCPSBC10.truck","2004NCPSBC11.truck","2004NCPSBC12.truck","2004NCPSBC13.truck","2004NCPSBC14.truck","2004NCPSBC15.truck","2004NCPSBC16.truck","CBM2004Freightliner.truck","2007C2cbm.truck","2007C2CBMV2HANDICAP.truck","2007C2CBMV2NEWYORK.truck","2007C2CBMV2ORIGINAL.truck","CBMMEVS43NEW.truck","2010CBMCE.truck","2010CBMCE2.truck","Bus33.truck","CBM2019ICRE-2.truck","CBMBBV.truck","2020BBVMrCRP.truck","ChrisBusModifications2020ThomasC2.truck","CBM1.truck","CBM2.truck","CBM3.truck","CBM4.truck","CBM5.truck","CBM6.truck","CBM7.truck","CBM8.truck","CBM9.truck","CBM10.truck","CBM2020ICRE-2.truck","ChrisBusModificationsAmTranFE-8.truck","BD642005CE200.truck","BD642006CE200.truck","BD642009CE300.truck","BD642010CE300.truck","BD642012CE300.truck","BD642013CE300.truck","CBM-BB3800-2.truck","CBM-BB3800-1.truck","Howland1.truck","Howland84.truck","Incandesent2005CE300MilfordSkin.truck","CBM2020CE300.truck","NBBCINTERNATIONAL2400-PublicR.truck","NBBCINTERNATIONAL2485-PublicR-1.1BugFix.truck","CBMMEVS41.truck","2004MEVS51-7.truck","2004MEVS64-7.truck","NBBC400-PUBLICR.truck","NBBC400-PUBLICR-3_WIM-I.truck","NBBC400-PUBLICR-3_WM-I.truck","NBBC400-PUBLICR-3_WIM.truck","NBBC400-PUBLICR-3_WM.truck","NBBC400-PUBLICR-4BugFixes_WIM-I.truck","NBBC400-PUBLICR-4BugFixes_WM-I.truck","NBBC400-PUBLICR-4BugFix_WIM.truck","NBBC400-PUBLICR-4BugFix_WM.truck","NBBC400-PUBLICR-5_WIM-I.truck","NBBC400-PUBLICR-5_WM-I.truck","NBBC500-PUBLICR-5_WIM.truck","NBBC500-PUBLICR-5_WM.truck","NBBC500-PUBLICR-6_WIM-I.truck","NBBC500-PUBLICR-6_WM-I.truck","NBBC500-PUBLICR-6_WIM.truck","NBBC500-PUBLICR-6_WM.truck","2004NCPSSC6.truck","2004NCPSSC7.truck","2004NCPSSC8.truck","2004NCPSSC9.truck","2004NCPSSC10.truck","2004NCPSSC11.truck","2004NCPSSC12.truck","2004NCPSSC13.truck","2004NCPSSC14.truck","2004NCPSSC15.truck","2004NCPSSC16.truck","LABS0867.truck","LABS0952.truck","LABS1251.truck","LABS1368.truck","LABS1369.truck","LABS1370.truck","LABS1371.truck","LABS1658.truck","LABS1660.truck","NCPSSC1.truck","NCPSSC2.truck","NCPSSC3.truck","NCPSSC4.truck","NCPSSC5.truck","NCPSSC25.truck","NCPSSC26.truck","2004CBM0427.truck","NoDistrict1.truck","NoDistrict2.truck","NoDistrict3.truck","NoDistrict4.truck","NoDistrict5.truck","NoDistrict6.truck","NoDistrict7.truck","NoDistrict8.truck","NoDistrict9.truck","NoDistrict10.truck","2006ICRE300ND.truck","2005CE200.truck","2006CE200.truck","2006ICRE300ND-2.truck","2000AmTranREBusHeritage66(HI)-3.truck","2000AmTranREBusHeritage66-3(HI) (NWSA).truck","2000AmTranREBusHeritage66-3(NWSA).truck","2000AmTranREBusHeritage66-3(WSA).truck","2000AmTranREBusHeritage66(HI).truck","2000AmTranREBusHeritage66(NWSA).truck","2000AmTranREBusHeritage66(WSA).truck","2000AmTranREBusHeritage66(HI) (NWSA).truck","Ryzen_H7_FPS_Incandesent.truck","Ryzen_X7000_NorthSchools50spare_Incandesent.truck","Ryzen_X7000_NorthSchools180-route_Incandesent.truck","Ryzen_X8000_NorthSchools50spare_Incandesent.truck","Ryzen_X8000_NorthSchools180-route_Incandesent.truck","RyzenFE.truck","Ryzen_X7000_FPS_Incandesent.truck","2019.truck","2001SEVS0138.truck","2002SEVS0215.truck","2004SEVS0311.truck","2004SEVS0312.truck","2004SEVS0314.truck","2004SEVS0316.truck","2004SEVS0318.truck","2004SEVS0319.truck","2004SEVS0321.truck","2004SEVS0327.truck","2004SEVS334.truck","2004SEVS0334.truck","2004SEVS0346.truck","CBM1999BLUEBIRDFREIGHTLINER.truck","2002CBMBB3800-0211.truck","CBMJASON2021ICCE.truck","JasonCBM2014CE300.truck","JasonCBM2014CE300mouthnoise.truck","ChrisBusModsMilfordExempted41Version2.truck","Bus66V4.truck","Bus66V4NOTWORKINGSTOP.truck","1999QISD9938.truck","1997GCPSCBM3800.truck","04BluebirdVISION.truck","CE-300.truck","1997BBGMC.truck","AISD1995ThomasMVPER.truck","AISD1996ThomasMVPER.truck","AISD1997ThomasMVPER.truck","AISD1999ThomasMVPER.truck","AISD2000ThomasMVPER.truck","AISD2001ThomasMVPER.truck","2003A3.truck","2006ICRE300FPS.truck","2006ICRE300.truck","NBBCFE400.truck","NBBCRE400.truck","2018C2-MPS-29-H.truck","2016ThomasC2-PASCO COUNTY SCHOOLS-61534.truck","LABS0535.truck","LABS9942.truck","2017ThomasC2-PASCO COUNTY SCHOOLS-61615.truck","2003QISD0314.truck","BCS0331.truck","NVS 44.truck","BCS0432.truck","BCS0323.truck","3888.truck","RidleySchoolDistrictBuses51(UPDATE).truck","RidleySchoolDistrictBus 53(UPDATE).truck","RidleySchoolDistrict Bus 75(UPDATE).truck","RidleySchoolDistrictBus 72(UPDATE).truck","2005QISD0532.truck","18-221.truck","54a9UID-2006Charger.truck","2018ICCE300-3724.truck","15-40.truck","2018BBV-SOUTHCAROLINA-0314.truck","2016ICCE10.truck","2017ICCE8.truck","MPS2020.21.truck","bus60.truck","69.truck","2021BluebirdVision736.truck","2020EISD2027.truck","2021EISD2130.truck","2020JasonTHS9.truck","MCPS140.truck","SwitzerCarty2016BluebirdVision732.truck","SwitzerCarty1999BB3800991.truck","SwitzerCarty2005CE200-180.truck","SwitzerCarty2012ThomasC2182.truck","SwitzerCarty2021ICCE300814.truck","SwitzerCarty2021ICCE300810.truck","SwitzerCarty14ICCE.truck","SwitzerCarty2008BluebirdVision530.truck","7700hybridkirsch.truck","2017DSS1726.truck","2018ICCE300-3725.truck","2018ICCE-BPS-B558.truck","SwitzerCarty2009BluebirdVision537.truck","SwitzerCarty2020ThomasC2.truck","SwitzerCartyC2213928.truck","SwitzerCartyICCE213706.truck","SwitzerCarty200200.truck","SwitzerCarty204801.truck","2019IC-CE -Bus 09-8.truck","2019IC-CE -Bus 13-02.truck","2019IC-CE -Bus 14-10.truck","2019IC-CE -Bus 18-41.truck","2019IC-CE -Bus 023-31.truck","2019IC-CE -Bus 23-01.truck","2019IC-CE -Bus 32-11.truck","2019IC-CE -Bus 49-38.truck","2019IC-CE -Bus 53-18.truck","2019IC-CE-Bus 09-10.truck","2019IC-CE-Bus 09-21.truck","2019IC-CE-Bus 20-93.truck","2019IC-CE-Bus 24-14.truck","99BBFS65EN.truck","2001QISD0117.truck","2002QISD0215.truck","2003QISD0311.truck","2003QISD0312.truck","2008BluebirdVision530.truck","2008BluebirdVision534.truck","2009BluebirdVision537.truck","2009BluebirdVision538.truck","2015BluebirdVision632.truck","2015BluebirdVision637.truck","2015BluebirdVision638.truck","2016BluebirdVision732.truck","2017BluebirdVision739.truck","2018BluebirdVision731.truck","2018BluebirdVision738.truck","2019BluebirdVision736.truck","2019BluebirdVision737.truck","2020BluebirdVision733.truck","1996GCPS9622.truck","1996GCPS9623.truck","1996GCPS9624.truck","1996GCPS9625.truck","1996GCPS9627.truck","1997GCPS9711.truck","3800D273.truck","C3M_IC_CE.truck","2011JTHS7.truck","2012JTHS10.truck","2013JTHS8.truck","2014JTHS1.truck","2016JTHS4.truck","2016JTHS6.truck","2017JTHS2.truck","2017JTHS5.truck","2018JTHS3.truck","2020JTHS9.truck","UPT1997BB3800.truck","UPT1997BBFS65.truck","UPT1997BluebirdChevy.truck","UPT1998BB3800.truck","UPT1998BBFS65.truck","UPT1998BluebirdChevy.truck","UPT1999BB3800.truck","UPT1999BBFS65.truck","UPT1999BluebirdChevy.truck","UPT2000BB3800.truck","UPT2000BBFS65.truck","UPT2001BB3800.truck","UPT2001BBFS65.truck","UPT2001BluebirdChevy.truck","UPT2002BB3800.truck","UPT2002BBFS65.truck","UPT2002BluebirdChevy.truck","UPT2003BluebirdChevy.truck","UPT2009BBV.truck","UPT2010BBV.truck","UPT2014BBV.truck","UPT2015BBV.truck","UPT2016BBV.truck","UPT2017BBV.truck","UPT2018BBV.truck","UPT2019BBV.truck","UPT2005ThomasHDX.truck","UPT2006ThomasHDX.truck","UPT2007ThomasHDX.truck","UPT2012ThomasHDX.truck","UPT2014ThomasHDX.truck","UPT2018ThomasHDX.truck","1994Thomas3800.truck","1995Thomas3800.truck","1996Thomas3800.truck","1997Thomas3800.truck","1998Thomas3800.truck","1999Thomas3800.truck","2000Thomas3800.truck","2001Thomas3800.truck","2002Thomas3800.truck","2003Thomas3800.truck","03-14.truck","03-41.truck","03-65.truck","03-75.truck","03-96.truck","04-42.truck","04-58.truck","04-61.truck","04-85.truck","04-99.truck","1997BBFS65.truck","1997BBFS651.truck","1997BBFS652.truck","1998BBFS65.truck","1998BBFS651.truck","1998BBFS652.truck","1999BBFS65.truck","1999BBFS651.truck","1999BBFS652.truck","1999BBFS653.truck","1997BlueBirdGMC3.truck","1997BlueBirdGMC4.truck","1998BlueBirdGMC.truck","1998BlueBirdGMC2.truck","1999BlueBirdGMC5.truck","2003BBChevy.truck","2003BBChevy7.truck","2003BBChevy8.truck","2003BBChevy9.truck","2004BBChevy10.truck","2006ICCE300Handicap.truck","2007ICCE300Handicap.truck","2009ICCE300Handicap.truck","2010ICCE300Handicap.truck","2011ICCE300Handicap.truck","2012ICCE300Handicap.truck","2014ICCE300Handicap.truck","2015ICCE300Handicap.truck","2017ICCE300Handicap.truck","2019ICCE300Handicap.truck","2005ICFE300.truck","2006ICFE300.truck","2007ICFE300.truck","2008ICFE300.truck","2009ICFE300.truck","2010ICFE300.truck","1997Crown9704.truck","1997Crown9707.truck","1997Crown9743.truck","1998Crown9823.truck","1998Crown9856.truck","1998Crown9875.truck","1999Crown9901.truck","1999Crown9908.truck","1999Crown9972.truck","1999Crown9999.truck","1995ThomasMVPER.truck","1996ThomasMVPER.truck","1997ThomasMVPER.truck","1999ThomasMVPER.truck","2000ThomasMVPER.truck","2001ThomasMVPER.truck","2001BB3300SBCV0186.truck","2002BB3300SBCV0217.truck","2002BB3300SBCV0234.truck","2003BB3300SBCV0334.truck","2003BB3300SBCV0371.truck","2004BB3300SBCV0443.truck","2004BB3300SBCV0475.truck","2005BB3300SBCV0560.truck","2005BB3300SBCV0591.truck","2005BB3300SBCV0598.truck","2006BB3300SBCV0692.truck","2000ThomasMVPERSCPS.truck","2001ThomasMVPERSCPS.truck","SCPS1995ThomasMVPER.truck","SCPS1996ThomasMVPER.truck","SCPS1997ThomasMVPER.truck","SCPS1998ThomasMVPER.truck","SCPS1999ThomasMVPER.truck","SCPS2000ThomasMVPER.truck","SCPS2001ThomasMVPER.truck","SCPS2002ThomasMVPER.truck","SCPS2003ThomasMVPER.truck","SCPS2004ThomasMVPER.truck","1996BBTC2000.truck","1997BBTC2000.truck","1997BBTC20009724.truck","1997BBTC20009725.truck","1998BBTC2000.truck","1999BBTC2000.truck","1999BBTC20009900.truck","1999BBTC20009940.truck","2007ThomasC2WB.truck","2007ThomasC2WY.truck","2007ThomasC2YY.truck","2008ThomasC2WB.truck","2008ThomasC2YB.truck","2008ThomasC2YY.truck","2009ThomasC2YB.truck","2010ThomasC2YB.truck","2013ThomasC2WY.truck","2014ThomasC2WY.truck","2014ThomasC2YY.truck","2015ThomasC2WB.truck","2015ThomasC2YB.truck","2015ThomasC2YY.truck","2018ThomasC2WB.truck","2018ThomasC2YB.truck","2019ThomasC2WB.truck","2019ThomasC2WY.truck","2019ThomasC2YB.truck","2019ThomasC2YY.truck","2005ICCE300NEW.truck","2006ICCE300NEW.truck","2007ICCE300NEW.truck","2008ICCE300NEW.truck","2009ICCE300NEW.truck","2010ICCE300NEW.truck","2011ICCE300NEW.truck","2012ICCE300NEW.truck","2013ICCE300NEW.truck","2014ICCE300NEW.truck","2015ICCE300NEW.truck","2016ICCE300NEW.truck","2017ICCE300NEW.truck","2018ICCE300NEW.truck","2019ICCE300NEW.truck","2005CE300.truck","2006CE300.truck","2007CE300.truck","2008CE300.truck","2009CE300.truck","2009CE3002.truck","2010CE300.truck","2011CE300.truck","2012CE300.truck","2013CE300.truck","2014CE300.truck","2015CE300.truck","2016CE300.truck","2016CE3002.truck","2017CE300.truck","2017CE3002.truck","2018CE300.truck","2018CE3002.truck","2019CE300.truck","2019CE3002.truck","2001ThomasFS65Cat7.truck","2001ThomasFS653126.truck","2002ThomasFS65Cat7.truck","2002ThomasFS653126.truck","2003ThomasFS65Cat7.truck","2003ThomasFS653126.truck","2004ThomasFS65Cat7.truck","2004ThomasFS653126.truck","2005ThomasFS65Cat7.truck","2005ThomasFS653126.truck","2003BBV.truck","2004BBV.truck","2005BBV.truck","2006BBV.truck","2007BBV.truck","2008BBV.truck","2009BBV.truck","2010BBV.truck","2011BBV.truck","2012BBV.truck","2013BBV.truck","2014BBV.truck","2015BBV.truck","2016BBV.truck","2017BBV.truck","2018BBV.truck","2019BBV.truck","1999BB3800991.truck","1999BB3800992.truck","2017ICCE.truck","2020ThomasC2.truck","LABS0044.truck","LABS0145.truck","LABS0246.truck","LABS0347.truck","LABS0448.truck","2012ThomasC2182.truck","Jason2005CE200-180.truck","Jason2006CE200-181.truck","MPSBUS9.truck","MPSBUS10.truck","MPSBUS11.truck","MPSBUS12.truck","MPSBUS28.truck","MPSBUS29.truck","MPSBUS41.truck","MPSBUS43.truck","MPSBUS45.truck","MPSBUS49.truck","MPSBUS53.truck","MPSBUS55.truck","MPSBUS57.truck","MPSBUS63.truck","14-1.truck","14-2.truck","ASD_131720.truck","ASD_132059.truck","ASD_132060.truck","ASD_132062.truck","ASD_132063.truck","ASD_132064.truck","ASD_132066.truck","ASD_132067.truck","ASD_132068.truck","ASD_132070.truck","ASD_132072.truck","ASD_132073.truck","ASD_132074.truck","ASD_132075.truck","ASD_132076.truck","ASD_132077.truck","ASD_132078.truck","ASD_JCJT6526.truck","ASD_JCJT6527.truck","2003ICFE300402.truck","2003ICFE300404.truck","2004ICFE300409.truck","2005ICFE300401.truck","2005ICFE300406.truck","2005ICFE300408.truck","2006ICFE300506.truck","2007ICFE300500.truck","2007ICFE300503.truck","2007ICFE300505.truck","2007ICFE300507.truck","2007ICFE300509.truck","2003ThomasSaf-t-linerHDX.truck","2004ThomasSaf-t-linerHDX.truck","2005ThomasSaf-t-linerHDX.truck","2006ThomasSaf-t-linerHDX.truck","2007ThomasSaf-t-linerHDX.truck","2008ThomasSaf-t-linerHDX.truck","2009ThomasSaf-t-linerHDX.truck","2010ThomasSaf-t-linerHDX.truck","2015ThomasSaf-t-linerHDX.truck","2018ThomasSaf-t-linerHDX.truck","2013ICCE300614.truck","2014ICCE300612.truck","2014ICCE300618.truck","2015ICCE300617.truck","2015ICCE300619.truck","2016ICCE300710.truck","2017ICCE300715.truck","2017ICCE300718.truck","2018ICCE300711.truck","2019ICCE300712.truck","2020ICCE300716.truck","2021ICCE300810.truck","2021ICCE300814.truck","UFN_TransitBus_0353.truck","UFN_TransitBus_0354.truck","UFN_TransitBus_0355.truck","UFN_TransitBus_0356.truck","UFN_TransitBus_0357.truck","UFN_TransitBus_0358.truck","UFN_TransitBus_0359.truck","UFN_TransitBus_0360.truck","UFN_TransitBus_0361.truck","UFN_TransitBus_0362.truck","UFN_TransitBus_0363.truck","T800HTator.truck","00P377.truck","04FLC120.truck","2004KWT600.truck","09KW900LB.truck","09KW900LA.truck","2010KWT660S.truck","2012Pinnacle.truck","2015Cascadia.truck","P378.truck","2006MackRawARI.truck","scota2hd6x4.truck","15VNL430N.truck","15VNL430.truck","13VolvoVN730Hi.truck","13VolvoVN670NEW.truck","VolvoVNL300W.truck","2013VN3002.truck","2013VN300.truck","06VolvoVN670.truck","Enzo.truck","2018ICCE300717.truck","2019ICCE300713.truck","10AcuraTSXT1.truck","2014-1455.truck","2014-1448.truck","2015-1591.truck","2018 - 813.truck","11CE300.truck","BMWe39Touring.truck","C3M_IC_BE.truck","C3M_IC_CE.truck","C3M_IC_CE_200.truck","C3M_BB_3800.truck","2013ICCE300614.truck","2021BluebirdVision737.truck","C3M_IC_CE_200.truck","2002BBFS65.truck","FSBus3.truck","FSBus9.truck","FSBus12.truck","FSBus13.truck","FSBus18.truck","FSBus19.truck","FSBus23.truck","FSBus29.truck","FSBus31.truck","FSBus43.truck","PB-21-16.truck","PB-29-17.truck","PB-5H-06.truck","2001 Toyota Chaser.truck","Murcielago.truck","vh-01fs65A.truck","2021 Thomas C2 bus #35.truck","2010QISD1036.truck","vh-b700.truck","2018ThomasC2ED8971.truck","2009ThomasC2ED8323.truck ","2018ThomasC2ED8971.truck","2006-ICCE-GCSNC-Bus-62.truck","vh-2005FS65898.truck","vh-1998FS65304.truck","55-2015BBIntWorkstar_S.truck","BUS 130.truck","BUS 061.truck","2016-581.truck","FST2015C2 10-12.truck","FST2013C2 10-11.truck","FST2013C3 10-15.truck","FST2013C2 10-16.truck","FST2012C2 10-02.truck","FST2011C2 10-10.truck","FST2011C2 10-19.truck","FST2008C2 10-17.truck","LABS1659.truck","FST2013C2 10-11.truck","T370_Rollback.truck","CCS0315.truck","APS1.truck","b740.truck","bbluebird.truck","bcapri.truck","2020ICCE-ED8405.truck","2020ICCE-ED8326.truck","2020ICCE-ED8321.truck","CCS2102.truck","2010ICCE-Clay.truck","vh-2003FS65987.truck","2016-ICCE-GCSNC-Bus-11.truck","2012-ICCE-GCSNC-Bus-3.truck","2010-ICCE-GCSNC-Bus-9.truck","08policetahoe.truck","08policetahoeslicktop.truck","2011FordTaruPolice.truck","06policeimpala.truck","06impalaunmarked.truck","12policecharger.truck","02CrownVicPolice.truck","02CrownVicPoliceALPR.truck","02CrownVicPolice Slicktop.truck","02CrownVicPolice Unmarked.truck","06chargerunmarked.truck","GHCSBUS009.truck","GHCSBUS024.truck","GHCSBUS011.truck","2016BBV_5.truck","GHCSBUS003.truck","2016BBV_9.truck","ROHRER 1826.truck","vh-2006FS651010.truck","1998ThomasVista[S].truck","gavrilmv2rGreenLightning.truck","accordsilvermissilenohood.truck","vh-02fs65130.truck","ROHRER 1722.truck","ROHRER 1603.truck","ROHRER 1604.truck","11DS3bl.truck","11Prius.truck","04_612.truck","bmwm1.truck","288gto.truck","diablogtr.truck","SV.truck","2015-IC-CE-GCSNC-Bus-6.truck","2017QISD1742.truck","2002PhotonRE488.truck","2003PhotonRE480.truck","2004PhotonRE485.truck","2005PhotonRE482.truck","2005PhotonRE489.truck","2006PhotonRE584.truck","2007PhotonRE580.truck","2009PhotonRE581.truck","notatruckC.truck","silvias13.truck","MVP500.truck","totallynotbakugan.truck","Iron Man 2020.truck","300ZX_RedBlade.truck","CIVIC_tuned.truck","RS6C8.truck","Chiron.truck","supraRZ.truck","cobra.truck","stingray.truck","GT15.truck","720s.truck","787b.truck","458_Italia_Green.truck","037.truck","mclarenf1gtr.truck","speedtail.truck","2021ThomasC2-SEMINOLE-J.truck","SSC_UA.truck","SLR.truck","MustangBoss.truck","11DS3vlt.truck","09Aveo.truck","vickpopo.truck","AudiR8.truck","86square.truck","07ClioRS.truck","LotusEspritFPSG.truck","mclaren12c.truck","Mercedes-Benz W126 SEC (coupe).truck","c63dtm.truck","aperta.truck","918.truck","812superfast.truck","s2000gt.truck","swedishmeatball.truck","dpivr.truck","2008_Ford_F250_2.truck","porsche935.truck","huracanperformante.truck","vulcan.truck","2015ICCE-1515ED.truck","MPSBUS32.truck","2015ICCE-1415J.truck","Bus25-04.truck","veneno.truck","2019TRRSD-264.truck","2018C2-MPS-21.truck","BLSD49.truck","BLSD24.truck","C-2007-89.truck","2019TRRSD-270.truck","C-2005-54.truck","C&F96.truck","2007C2SLT150.truck","vh-2006FS65SC.truck","m1ahg.truck","Multipla.truck","FiatPuntoAbarthS1600.truck","Maluch.truck","ducato.truck","T5.truck","BentaygaStartech.truck","CNTNNTL.truck","X5M.truck","SuperB.truck","Taycan.truck","PassatB8.truck","miurasv.truck","m1ahgup.truck","2018ICCE300717.truck","2017gt.truck","swedishmeatball.truck","AstonMartin_DB5.truck","san-ju-yon.truck","G63Crazy.truck","ChrisBusModsMilfordExempted41Version2.truck","CSDS14CE1400.truck","CSDS18CE1800.truck","CSDS15CE1500.truck","CSDS21C22100.truck","2006-IC-CE GCSNC-Bus-19.truck","Mazda6MPS.truck","WRXSTI.truck","2007HondaCivicByTheMrCrasher99.truck","LBNSX.truck","notatoyota.truck","92 ford offroad wrecker.truck","2020ICCE-ED8324.truck","2018ICCE300711.truck","2015BluebirdVision638.truck","2002PhotonRE488.truck","2009PhotonRE581.truck","2006PhotonRE584.truck","2018BBV-FS-218.truck","2018BBV-FS-118.truck","2002_A3.truck","2010EF.truck","2015ICCE619.truck","2001ABL01-23.truck","1998Bluebird3800377.truck","75fordwrecker1.truck","newfordwrecker1.truck","AventadorJ.truck","Mazda6MPSTuned.truck","Culinan.truck","G5004x4.truck","LamboUrus.truck","CE300.truck","S64SPORT.truck","2010ThomasC2HandyFS949.truck","2010ThomasC2HandyFS947.truck","2016ThomasC2-FS-939.truck","2016ThomasC2-FS-933.truck","2016ThomasC2-FS-937.truck","2010ThomasC2HandyFS948.truck","2014BBV-FS-214.truck","2017BBV_SERSD-20.truck","2000BB3800-FS-100.truck","2017BBV_SERSD_23.truck","2004BB3800-SERSD-254.truck","2016ThomasC2-FS-934.truck","2015ICCE-FS-315.truck","2017BBV_SERSD-25.truck","2004BB3800-SERSD-251.truck","2016ThomasC2-FS-938.truck","amcgremlin.truck","03BoraFPS.truck","R34vspec2custom.truck","dmc12CyberMidnight.truck","2002BBFL70-FS-102.truck","2015ICCE-FS-215.truck","2016ThomasC2-FS-940.truck","CE300MT463.truck","2001BB3800-SERSD-252.truck","1995-Ford-B700.truck","eclipse.truck","1997RX-7closed.truck","SLSAMG.truck","M5.truck","4C.truck","Stelvio.truck","AlfaTZ3.truck","notabmw.truck","911rsr.truck","M2.truck","amgt63.truck","clkgtr.truck","a45.truck","diablosv.truck","a1523UIDcountach_twilight-ray.truck","veyron164.truck","sianFKP37.truck","mcsennagtr.truck","Performante.truck","c7r.truck","accordsilvermissile.truck","m3_christmas.truck","m3_stock.truck","fxxkevowhite.truck","fxxkevo.truck","gtgt1.truck","bt62.truck","venomGT.truck","vantageAMR.truck","2017gtwing.truck","nsxgt3.truck","r8lmsultra.truck","gt2rs.truck","G636x6.truck","gallardogt3.truck","aventadorsvj.truck","Velar.truck","SVR.truck","FSPSBUS 09.truck","CRSD 2016 Bluebird Vision CR03.truck","CRSD 2021 CR04.truck","1997Bluebird3800375.truck","1997Bluebird3800374.truck","1997Bluebird3800373.truck","1998Bluebird3800372.truck","1997Bluebird3800371.truck","1997Bluebird3800370.truck","BBV974.truck","S1800-DT408.truck","Race_Touareg .truck","1987-Chevrolet-Diesel-60.truck","2009BBA3FE-WCSD-104.truck","2009BBA3FE-KSD-101.truck","2009BBA3FE-WCSD-100.truck","2014 EFX.truck","1995-International-3800-DT466-activity.truck","2016BBV_2.truck","S1800-DT466.truck","f150goodbogger.truck","bigbogger79.truck","ACX-NC.TRUCK","NCPS-2016BBV-206.truck","2009BBA3FE-WCSD-103.truck","2009BBA3FE-WCSD-101.truck","FSPSBUS01.truck","FSPSBUS05.truck","FSPSBUS03.truck","FSPSBUS 07.truck","FerrariCaliforniaO.truck","11PriusFPS.truck","koenigccxr.truck","BBV986.truck","BBV987.truck","FSPSBUS 09.truck","CRSD 2016 Bluebird Vision CR03.truck","CRSD 2021 CR04.truck","49fordbeater.truck","ROHRER 1600.truck","GHCSBUS021.truck","2010ICCE-FL.truck","SMFCS33.truck","DodgeChargerPolice.truck","mustang_zephyr.truck","C4Picasso.truck","599xx.truck","LaFerrari.truck","Jag.truck","ISF.truck","GSF.truck","RCF.truck","E63s.truck","BRZP.truck","Jason2017VisionNCarolina677.truck","Maybach.truck","genesis.truck","GHCSBUS010.truck","DSS201472.truck","A&S TRAN (032).truck","ttrs.truck","96ER2090.truck","2019ThomasHDX-SEMINOLE-61811.truck","2012ThomasHDX-SEMINOLE-61210.truck","2013BluebirdVision61308.truck","2012BluebirdVision61206.truck","1999A&STC2000Bus5.truck","D40LF.truck","2013ICCE300614.truck","2000BB3800-FS-200.truck","C-2007-86.truck ","VelosterTurbo.truck","C63AMG.truck","gallardo.truck","gavrilgv5_SSR.truck","2006 IC CE N.Y.P.D.truck","carreragt.truck","boxster.truck","159.truck","C-2008-18.truck","2014BBV-FS-114.truck","HondaIntegraR.truck","modena.truck","burnside_derby.truck","2021ICCET2.truck","1729.truck","VWGolfGTiMK2.truck","2010 BB AA Wisconsin.truck","gt2club.truck","2021 C2 - 2125.truck","2021 C2 - 2127.truck","gavrilgv5_BBS.truck","vh-2005FS65898.truck"
}; 
streamValidation bannedCars(@streamFilterList, BLACKLIST);


// Our main function (never remove this, even when it does nothing useful!)
void main()
{
	server.Log("The script file is working :D");
	server.setCallback("playerAdded", "authWarning_playerAdded", null);
}

void authWarning_playerAdded(int uid)
{
	if((server.getUserAuthRaw(uid)&(AUTH_ADMIN|AUTH_MOD))>0)
		server.say(server.getUserName(uid)+" is a server "+server.getUserAuth(uid), TO_ALL, FROM_HOST);
}